"""
Check accessing Exception.message
"""
# pylint: disable=import-error, no-absolute-import, broad-except

from unknown import ExtensionException
__revision__ = 0

class SubException(IndexError):
    """ empty """

_ = IndexError("test").message # [exception-message-attribute]
_ = ZeroDivisionError("error").message # [exception-message-attribute]
_ = ExtensionException("error").message
_ = SubException("error").message # [exception-message-attribute]

try:
    raise Exception('e')
except Exception as exception:
    _ = exception.message # [exception-message-attribute]
    del exception.message # [exception-message-attribute]
    exception.message += 'hello world' # [exception-message-attribute]
    exception.message = 'hello world'


class CompatException(Exception):
    """An exception which should work on py2 and py3."""

    def __init__(self, message=''):
        super(CompatException, self).__init__()
        self.message = message

    def __repr__(self):
        result = 'CompatException %s' % self.message
        return result.encode('utf-8')


try:
    raise CompatException('message here')
except CompatException as error:
    _ = error.message



"""Test for unsupported-binary-operation."""
# pylint: disable=missing-docstring,too-few-public-methods,pointless-statement
# pylint: disable=expression-not-assigned, invalid-name


import collections


1 + "a" # [unsupported-binary-operation]
1 - [] # [unsupported-binary-operation]
1 * {} # [unsupported-binary-operation]
1 / collections # [unsupported-binary-operation]
1 ** (lambda x: x) # [unsupported-binary-operation]
{} * {} # [unsupported-binary-operation]
{} - {} # [unsupported-binary-operation]
{} | {} # [unsupported-binary-operation]
{} >> {} # [unsupported-binary-operation]
[] + () # [unsupported-binary-operation]
() + [] # [unsupported-binary-operation]
[] * 2.0 # [unsupported-binary-operation]
() * 2.0 # [unsupported-binary-operation]
2.0 >> 2.0 # [unsupported-binary-operation]
class A(object):
    pass
class B(object):
    pass
A() + B() # [unsupported-binary-operation]
class A1(object):
    def __add__(self, other):
        return NotImplemented

A1() + A1() # [unsupported-binary-operation]

class A2(object):
    def __add__(self, other):
        return NotImplemented
class B2(object):
    def __radd__(self, other):
        return NotImplemented
A2() + B2() # [unsupported-binary-operation]

class Parent(object):
    pass
class Child(Parent):
    def __add__(self, other):
        return NotImplemented
Child() + Parent() # [unsupported-binary-operation]
class A3(object):
    def __add__(self, other):
        return NotImplemented
class B3(A3):
    def __radd__(self, other):
        return NotImplemented
A3() + B3() # [unsupported-binary-operation]
# Augmented
FFF = 1
FFF += A() # [unsupported-binary-operation]
TTT = 1
TTT += [] # [unsupported-binary-operation]


# Don't emit for this case since we don't know what unknown is.
from unknown import Unknown
class Base(Unknown):
    pass
Base() * 23



# pylint: disable=literal-comparison,missing-docstring

X = ''
Y = 'test'

if X is '':  # [compare-to-empty-string]
    pass

if Y is not "":  # [compare-to-empty-string]
    pass

if X == "":  # [compare-to-empty-string]
    pass

if Y != '':  # [compare-to-empty-string]
    pass



"""Warnings for using open() with an invalid mode string."""

NAME = "foo.bar"
open(NAME, "wb")
open(NAME, "w")
open(NAME, "rb")
open(NAME, "x")
open(NAME, "br")
open(NAME, "+r")
open(NAME, "xb")
open(NAME, "rwx")  # [bad-open-mode]
open(NAME, "rr")  # [bad-open-mode]
open(NAME, "+")  # [bad-open-mode]
open(NAME, "xw")  # [bad-open-mode]
open(NAME, "ab+")
open(NAME, "a+b")
open(NAME, "+ab")
open(NAME, "+rUb")
open(NAME, "x+b")
open(NAME, "Ua")  # [bad-open-mode]
open(NAME, "Ur++")  # [bad-open-mode]
open(NAME, "Ut")
open(NAME, "Ubr")



"""Check exceeding negations in boolean expressions trigger warnings"""

# pylint: disable=singleton-comparison,too-many-branches,too-few-public-methods,undefined-variable
# pylint: disable=literal-comparison
def unneeded_not():
    """This is not ok
    """
    bool_var = True
    someint = 2
    if not not bool_var:  # [unneeded-not]
        pass
    if not someint == 1:  # [unneeded-not]
        pass
    if not someint != 1:  # [unneeded-not]
        pass
    if not someint < 1:  # [unneeded-not]
        pass
    if not someint > 1:  # [unneeded-not]
        pass
    if not someint <= 1:  # [unneeded-not]
        pass
    if not someint >= 1:  # [unneeded-not]
        pass
    if not not someint:  # [unneeded-not]
        pass
    if not bool_var == True:  # [unneeded-not]
        pass
    if not bool_var == False:  # [unneeded-not]
        pass
    if not bool_var != True:  # [unneeded-not]
        pass
    if not True == True:  # [unneeded-not]
        pass
    if not 2 in [3, 4]:  # [unneeded-not]
        pass
    if not someint is 'test':  # [unneeded-not]
        pass


def tolerated_statements():
    """This is ok"""
    bool_var = True
    someint = 2
    if not(bool_var == False and someint == 1):
        pass
    if 2 not in [3, 4]:
        pass
    if not someint == bool_var == 2:
        pass
    if not 2 <= someint < 3 < 4:
        pass
    if not set('bar') <= set('foobaz'):
        pass
    if not set(something) <= 3:
        pass
    if not frozenset(something) <= 3:
        pass


class Klass(object):
    """This is also ok"""
    def __ne__(self, other):
        return not self == other



# pylint: disable=C0111,R0903
"""#3291"""
from __future__ import print_function

class Myarray(object):
    def __init__(self, array):
        self.array = array

    def __mul__(self, val):
        return Myarray(val)

    def astype(self):
        return "ASTYPE", self

def randint(maximum):
    if maximum is not None:
        return Myarray([1, 2, 3]) * 2

    return int(5)

print(randint(1).astype()) # we don't wan't an error for astype access



"""Check invalid value returned by __len__ """

# pylint: disable=too-few-public-methods,missing-docstring,no-self-use,import-error
import sys

import six

from missing import Missing


class FirstGoodLen(object):
    """__len__ returns <type 'int'>"""

    def __len__(self):
        return 0


class SecondGoodLen(object):
    """__len__ returns <type 'long'>"""

    def __len__(self):
        return sys.maxsize + 1


class LenMetaclass(type):
    def __len__(cls):
        return 1


@six.add_metaclass(LenMetaclass)
class ThirdGoodLen(object):
    """Length through the metaclass."""


class FirstBadLen(object):
    """ __len__ returns a negative integer """

    def __len__(self):  # [invalid-length-returned]
        return -1


class SecondBadLen(object):
    """ __len__ returns non-int """

    def __len__(self):  # [invalid-length-returned]
        return 3.0


class ThirdBadLen(object):
    """ __len__ returns node which does not have 'value' in AST """

    def __len__(self):  # [invalid-length-returned]
        return lambda: 3


class AmbigousLen(object):
    """ Uninferable return value """
    __len__ = lambda self: Missing


class AnotherAmbiguousLen(object):
    """Potential uninferable return value"""
    def __len__(self):
        return int(Missing)



# pylint: disable=too-few-public-methods,print-statement
"""check method hidding ancestor attribute
"""
from __future__ import print_function

class Abcd(object):
    """dummy"""
    def __init__(self):
        self.abcd = 1

class Cdef(Abcd):
    """dummy"""
    def abcd(self): # [method-hidden]
        """test
        """
        print(self)



"""Tests for redefining builtins."""
from __future__ import print_function

def function():
    """Redefined local."""
    type = 1  # [redefined-builtin]
    print(type)

# pylint:disable=invalid-name
map = {}  # [redefined-builtin]



"""Test warnings about access to undefined variables
for various Python 3 constructs. """
# pylint: disable=too-few-public-methods, no-init, no-self-use
# pylint: disable=wrong-import-position, invalid-metaclass
class Undefined:
    """ test various annotation problems. """

    def test(self)->Undefined: # [undefined-variable]
        """ used Undefined, which is Undefined in this scope. """

    Undefined = True

    def test1(self)->Undefined:
        """ This Undefined exists at local scope. """

    def test2(self):
        """ This should not emit. """
        def func()->Undefined:
            """ empty """
            return
        return func


class Undefined1:
    """ Other annotation problems. """

    Undef = 42
    ABC = 42

    class InnerScope:
        """ Test inner scope definition. """

        def test_undefined(self)->Undef: # [undefined-variable]
            """ Looking at a higher scope is impossible. """

        def test1(self)->ABC: # [undefined-variable]
            """ Triggers undefined-variable. """


class FalsePositive342(object):
    # pylint: disable=line-too-long
    """ Fix some false positives found in
    https://bitbucket.org/logilab/pylint/issue/342/spurious-undefined-variable-for-class
    """

    top = 42

    def test_good(self, bac: top):
        """ top is defined at this moment. """

    def test_bad(self, bac: trop): # [undefined-variable]
        """ trop is undefined at this moment. """

    def test_bad1(self, *args: trop1): # [undefined-variable]
        """ trop1 is undefined at this moment. """

    def test_bad2(self, **bac: trop2): # [undefined-variable]
        """ trop2 is undefined at this moment. """

from abc import ABCMeta

class Bad(metaclass=ABCMet): # [undefined-variable]
    """ Notice the typo """

class SecondBad(metaclass=ab.ABCMeta): # [undefined-variable]
    """ Notice the `ab` module. """

class Good(metaclass=int):
    """ int is not a proper metaclass, but it is defined. """

class SecondGood(metaclass=Good):
    """ empty """

class ThirdGood(metaclass=ABCMeta):
    """ empty """

class FourthGood(ThirdGood):
    """ This should not trigger anything. """

# The following used to raise used-before-assignment
# pylint: disable=missing-docstring, multiple-statements
def used_before_assignment(*, arg): return arg + 1
